#!/bin/lua

--
-- INFO: to check/test a specific configuration, call this skript with appropiate values:
-- lua -e "platform='SimSpark'; scheme='RC17'; robotName='nao96'; bodyID=''; headID=''" Config/config.lua 
--

--[[
--  TODO:
--    - protect changes of the config table in (sub-)files
--    - set config vars via argument (eg. it could be possible to call the config file for debugging/check/test reasons?!)
--]]

-- imports
require "pl"

-- config vars
local configDir = configDir or (arg and arg[0]:match("@?(.*/)") or "")
local platform = platform or ""
local scheme = scheme or ""
local robotName = robotName or ""
local bodyID = bodyID or ""
local headID = headID or ""

-- the directories where the (sub)configs are located
local dirs = {
  configDir .. "general/",
  configDir .. "platform/"     .. platform, 
  configDir .. "scheme/"       .. scheme, 
  configDir .. "robots/"       .. robotName, 
  configDir .. "robots_bodies/".. bodyID, 
  configDir .. "robot_heads/"  .. headID, 
  configDir .. "private/",
}

-- the final config table
local config = {}

---
--- function definitions
---

--- print function for debugging
function luaprint(...)
    print("lua >", ...)
end

function printConfig()
  io.write("\nconfig = ")
  dumpTable(config or {})
  io.write("\nprivate = ")
  dumpTable(private or {})
end

function getField(f)
  --luaprint(f)
  local v = _G    -- start with the table of globals
  for w in string.gmatch(f, "[%w_:]+") do
    --luaprint(v, w)
    if type(v[w]) ~= "nil" then
      v = v[w]
    else
      return nil
    end
  end
  --luaprint(v)
  return v
end

function setField (f, v)
  --luaprint(f .. " = " .. tostring(v) .." (" .. type(v) .. ")")
    local t = _G    -- start with the table of globals
    for w, d in string.gmatch(f, "([%w_:]+)(.?)") do
      if d == "." then      -- not last field?
        t[w] = t[w] or {}   -- create table if absent
        t = t[w]            -- get the table
      else                  -- last field
        t[w] = v            -- do the assignment
      end
    end
  end

function dumpTable(t)
  require 'pl.pretty'.dump(t)
end

function include(configfile)
  -- define the metatable for the config file
  local env = setmetatable({}, {
    __index=function(t,k)
      if _G[k] then return _G[k]
      elseif config[k] then return config[k]
      else return t[k]
      end
    end
  })
  --TODO: use "pcall"?!
  -- load and execute file
  assert(loadfile(configfile, 't', env))()
  -- only return the vars defined by the config file
  setmetatable(env, nil)
  return env
end

function getFields(t)
  --
  local keys = {}
  local n = 0
  local group = getField(t)
  --
  map = function(tt,p)
    for k,v in pairs(tt) do
      if type(tt[k]) == "table" then
        map(tt[k], p .. k .. ".")
      else
        n=n+1
        keys[n] = p .. k
      end
    end
  end
  -- 
  if type(group) == "table" then
    map(group,"")
  end

  return keys
end

function savePrivate()
  path = dirs[#dirs]
  if private then
    for k,v in pairs(private) do
      local file, e = io.open(path .. "/" .. k .. ".lua", "w");
      if not file then
        print("[ERROR] Writing private file (" .. path .. "/" .. k .. ".lua" .. "):" .. e)
      else
        for h,j in pairs(v) do
          if type(j) == "table" then
            file:write("_ENV['" .. h .. "'] = " .. pretty.write(j) .. "\n")
          else
            file:write("_ENV['" .. h .. "'] = " .. j .. "\n")
          end
        end -- for h,j
        file:close()
      end -- if not file
    end -- for k,v
  else
    --luaprint("nothing to save")
  end
end

---
--- Main
---

for i,v in ipairs(dirs) do
  -- ignore "access errors"
  if not pcall(function()
  --
    -- TODO: does this only work on linux?!?
    --       maybe: something like "require 'lfs'.dir(v)"
    local p = io.popen('find "'..v..'" -type f')
    for file in p:lines() do
      -- only lua files
      if string.lower(string.sub(file,-4)) == ".lua" then
        local name = file
        local key = string.match(file, "([^/]+).lua") --string.sub(file,1,-5)
        local status, m = pcall(include, name)
        if status then
          if config[key] then
            --join tables
            for k,v in pairs(m) do config[key][k] = v end
          else
            config[key] = m
          end
          print("[INFO] loaded " .. name)
        else
          print("[WARNING] Could not load config file: " .. name .. "\n\t" .. tostring(m))
        end
      end
    end
  end) then
    print("[WARNING] Could not load config files from: " .. v)
  end
end

-- make local config global
_G.config = config

-- only print config if requested
if arg and arg[1] == "print" then
  printConfig()
end


-- DEBUG
--print(getField("config.x"))
--print(getField("config.CameraInfoBottom.openingAngleDiagonal"))

--setField("config.CameraMatrixOffset.body_rot.x", 0.0)
--print(getField("config.CameraMatrixOffset.body_rot.x"))

--setField("config.TeamCommReceiver.monotonicTimestampCheck", 1)
--print(getField("config.TeamCommReceiver.monotonicTimestampCheck"))

--setField("config.CameraMatrixOffset.cam_rotnaoth::CameraInfo::Bottom.x", 0.00462276)
--print(getField("config.CameraMatrixOffset.cam_rotnaoth::CameraInfo::Bottom.x"))
--dumpTable(getField("config.CameraMatrixOffset"))


--setField("private.CameraMatrixOffset.body_rot.x", 0.0)
--setField("private.TeamCommReceiver.monotonicTimestampCheck", 1)
--setField("private.CameraMatrixOffset.cam_rotnaoth::CameraInfo::Bottom.x", 0.00462276)

--t = include("/home/philipp/Development/Lua/test.lua")
--print(t)
--dumpTable(t)
--t = load('do local _={["cam_rotnaoth::CameraInfo::Bottom"]={x=0.0046227600000000001},body_rot={x=0}};return _;end')
--y = t()
--print(y)
--dumpTable(y)

--savePrivate()

